<?php

namespace Drupal\jsonapi_access\EventSubscriber;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityAccessControlHandlerInterface;
use Drupal\jsonapi\CacheableResourceResponse;
use Drupal\jsonapi\JsonApiResource\JsonApiDocumentTopLevel;
use Drupal\jsonapi\Query\Filter;
use Drupal\jsonapi\ResourceResponse;
use Drupal\jsonapi_extras\ResourceType\ConfigurableResourceType;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Class ResponseSubscriber.
 */
class ResponseSubscriber implements EventSubscriberInterface {

  /**
   * @var \Drupal\Component\DependencyInjection\ContainerInterface
   */
  private ContainerInterface $container;

  public function __construct(ContainerInterface $container) {
    $this->container = $container;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events[KernelEvents::RESPONSE] = ['onResponse', 512];

    return $events;
  }

  /**
   * This method is called the KernelEvents::RESPONSE event is dispatched.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The filter event.
   */
  public function onResponse(ResponseEvent $event): void {
    if (!$event->isMainRequest()) {
      return;
    }
    $request = $event->getRequest();
    $response = $event->getResponse();
    $include_access = explode(',', $request->get('jsonapi_access', ''));
    if (!$response instanceof ResourceResponse || !in_array('create', $include_access)) {
      return;
    }
    $route_match = $this->container->get('current_route_match');
    $resource_type = $route_match->getParameter('resource_type');
    if (!$resource_type instanceof ConfigurableResourceType) {
      return;
    }
    $em = $this->container->get('entity_type.manager');
    $access = $em->getAccessControlHandler($resource_type->getEntityTypeId());
    if (!$access instanceof EntityAccessControlHandlerInterface) {
      return;
    }
    $query = $request->query->all();
    $context[Filter::KEY_NAME] = $query[Filter::KEY_NAME] ?? [];
    $meta['access'] = [];
    $accessResult = $access->createAccess($resource_type->getBundle(), NULL, $context, TRUE);
    if ($accessResult->isAllowed()) {
      $meta['access'][] = 'create';
    }
    if ($response instanceof CacheableResourceResponse) {
      $cache = new CacheableMetadata();
      $cache->addCacheTags([$resource_type->getEntityTypeId() . '_jsonapi_access']);
      $cache->addCacheableDependency($accessResult);
      $response->addCacheableDependency($cache);
    }
    $event->setResponse($this->createResponse($response, $meta));
  }

  /**
   * Re-create json response with meta access.
   *
   * @param \Drupal\jsonapi\ResourceResponse $response
   *   Original response object.
   *
   * @param array $meta
   *   Additional meta information.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   Response with additional parameters or original if not applicable.
   */
  protected function createResponse(ResourceResponse $response, array $meta): Response {
    $responseData = $response->getResponseData();
    if (!$responseData instanceof JsonApiDocumentTopLevel) {
      return $response;
    }
    $data = $responseData->getData();
    $includes = $responseData->getIncludes();
    $links = $responseData->getLinks();
    $meta = $meta + $responseData->getMeta();

    $class = get_class($response);
    $newResponse = new $class(
      new JsonApiDocumentTopLevel($data, $includes, $links, $meta),
      $response->getStatusCode(),
      $response->headers->all()
    );
    if ($response instanceof CacheableResourceResponse && $newResponse instanceof CacheableResourceResponse) {
      $cacheMeta = $response->getCacheableMetadata();
      $cacheMeta->addCacheContexts(['url.query_args:jsonapi_access']);
      $newResponse->addCacheableDependency($cacheMeta);
    }
    return $newResponse;
  }

}
